"""
Module with some simple channel estimation functions.

Note: See also :mod:`reference_signals.channel_estimation`.
"""

import math

import numpy as np


def compute_ls_estimation(Y_p: np.ndarray, s: np.ndarray) -> np.ndarray:
    """
    Compute the channel estimation using the LS estimator.

    The estimated channel is for a SIMO channel (multiple receive antennas)
    where the channel is assumed constant during the transmission of all the
    pilots.

    The formula comes from the "Performance analysis of block and comb type
    channel estimation for massive MIMO systems (Gábor Fodor at all) - 2014"
    paper.

    Parameters
    ----------
    Y_p : np.ndarray
        The received symbols. Dimension must be either `Nr x num_pilots` or
        `num_realizations x Nr x num_pilots`
    s : np.ndarray
        The transmitted pilots. Dimension must be either `Nt x num_pilots` or
        `num_realizations x Nt x num_pilots`

    Returns
    -------
    np.ndarray
        The estimated channel using the LS algorithm.
        Dimension: either `Nr x Nt` or `num_realizations x Nr x Nt`
    """
    if Y_p.ndim == 2:
        assert (s.ndim == 2)
        return Y_p @ s.T.conj() @ np.linalg.inv(s @ s.conj().T)

    num_realizations, Nr, num_pilots = Y_p.shape

    if s.ndim == 2:
        # use the same pilots in all channel realizations
        Nt = s.shape[0]
        out = np.empty((num_realizations, Nr, Nt),
                       dtype=np.common_type(Y_p, s))
        for i in range(num_realizations):
            out[i] = compute_ls_estimation(Y_p[i], s)
    else:
        # use different pilots in the different channel realizations
        assert (s.shape[0] == num_realizations)
        Nt = s.shape[1]
        out = np.empty((num_realizations, Nr, Nt),
                       dtype=np.common_type(Y_p, s))
        for i in range(num_realizations):
            out[i] = compute_ls_estimation(Y_p[i], s[i])

    return out


def compute_theoretical_ls_MSE(Nr: int, noise_power: float, alpha: float,
                               pilot_power: float, num_pilots: int) -> float:
    """
    Compute the theoretical MSE for the LS channel estimator.

    The estimated channel is for a SIMO channel (multiple receive antennas)
    where the channel is assumed constant during the transmission of all the
    pilots.

    The formula comes from the "Performance analysis of block and comb type
    channel estimation for massive MIMO systems (Gábor Fodor at all) - 2014"
    paper.

    Parameters
    ----------
    Nr : int
        The number of receive antennas.
    noise_power : float
        The noise power.
    alpha : float
        The linear path loss.
    pilot_power : float
        The pilot power.
    num_pilots : int
        The  number of pilots.

    Returns
    -------
    float
        The computed MSE.

    @see compute_ls_estimation
    """
    return Nr * noise_power / ((alpha**2) * pilot_power * num_pilots)


def compute_mmse_estimation(Y_p: np.ndarray, s: np.ndarray, noise_power: float,
                            C: np.ndarray) -> np.ndarray:
    """
    Compute the channel estimation using the MMSE estimator.

    The estimated channel is for a SIMO channel (multiple receive antennas)
    where the channel is assumed constant during the transmission of all the
    pilots.

    Note: If there is path loss it should be accounted in the channel covariance
    matrix.

    The formula comes from the "Performance analysis of block and comb type
    channel estimation for massive MIMO systems (Gábor Fodor at all) - 2014"
    paper.

    Parameters
    ----------
    Y_p : np.ndarray
        The received symbols. Dimension must be either `Nr x num_pilots` or
        `num_realizations x Nr x num_pilots`
    s : np.ndarray
        The transmitted pilots. Dimension must be either `Nt x num_pilots` or
        `num_realizations x Nt x num_pilots`
    noise_power : float
        The noise power.
    C : np.ndarray
        The channel covariance matrix (reveive antennas)

    Returns
    -------
    np.ndarray
        The estimated channel using the MMSE algorithm.
        Dimension: either `Nr x Nt` or `num_realizations x Nr x Nt`
    """
    if Y_p.ndim == 2:
        assert (s.ndim == 2)
        Nt = s.shape[0]
        assert (Nt == 1)
        Nr, num_pilots = Y_p.shape

        # The model is Y_vec = S h + N
        Y_vec = np.reshape(Y_p, (Nr * num_pilots, 1), order="F")
        S = np.kron(s.T, np.eye(Nr))

        I_Nr = np.eye(Nr)

        pilot_power_times_num_pilots = s @ s.T.conj()

        return (
            np.linalg.inv(noise_power * I_Nr + num_pilots * C) @ C
            @ S.T.conj()) @ Y_vec / pilot_power_times_num_pilots * num_pilots

    num_realizations, Nr, num_pilots = Y_p.shape

    # Case where Y_p has 3 dimensions
    if s.ndim == 2:
        # use the same pilots in all channel realizations
        Nt = s.shape[0]
        assert (Nt == 1)
        out = np.empty((num_realizations, Nr, Nt),
                       dtype=np.common_type(Y_p, s))
        for i in range(num_realizations):
            out[i] = compute_mmse_estimation(Y_p[i], s, noise_power, C)
    else:
        # use different pilots in the different channel realizations
        assert (s.shape[0] == num_realizations)
        Nt = s.shape[1]
        assert (Nt == 1)
        out = np.empty((num_realizations, Nr, Nt),
                       dtype=np.common_type(Y_p, s))
        for i in range(num_realizations):
            out[i] = compute_mmse_estimation(Y_p[i], s[i], noise_power, C)

    return out


def compute_theoretical_mmse_MSE(Nr, noise_power, alpha, pilot_power,
                                 num_pilots, C):
    """
    Compute the theoretical MSE for the MMSE channel estimator.

    The estimated channel is for a SIMO channel (multiple receive antennas)
    where the channel is assumed constant during the transmission of all the
    pilots.

    The formula comes from the "Performance analysis of block and comb type
    channel estimation for massive MIMO systems (Gábor Fodor at all) - 2014"
    paper.

    Parameters
    ----------
    Nr : int
        The number of receive antennas.
    noise_power : float
        The noise power.
    alpha : float
        The linear path loss.
    pilot_power : float
        The pilot power.
    num_pilots : int
        The  number of pilots.
    C : np.ndarray
        The channel covariance matrix (reveive antennas)

    Returns
    -------
    float
        The computed MSE.

    @see compute_mmse_estimation
    """
    return np.trace(C @ np.linalg.inv(
        np.eye(Nr) + alpha**2 * pilot_power * num_pilots / noise_power * C))



#!/usr/bin/env python
"""
Module implementing OFDM modulation and demodulation.
"""

import math
from typing import Optional, Tuple

import numpy as np

from ..channels import fading

__all__ = ['OFDM', 'OfdmOneTapEqualizer']


class OFDM:
    """
    OFDM class.
    """
    def __init__(self,
                 fft_size: int,
                 cp_size: int,
                 num_used_subcarriers: Optional[int] = None) -> None:
        """
        Initialize the OFDM object.

        Parameters
        ----------
        fft_size : int
            Size of the FFT and IFFT used by the OFDM class.
        cp_size : int
            Size of the cyclic prefix (in samples).
        num_used_subcarriers : int, optional
            Number of used subcarriers. Must be greater than or equal to 2
            and lower than or equal to fft_size. If not provided, fft_size
            will be used

        Returns
        -------
        OFDM

        Raises
        ------
        ValueError
            If the any of the parameters are invalid."""
        self.fft_size: int = 0
        self.cp_size: int = 0
        self.num_used_subcarriers: int = 0

        self.set_parameters(fft_size, cp_size, num_used_subcarriers)

    def set_parameters(self,
                       fft_size: int,
                       cp_size: int,
                       num_used_subcarriers: Optional[int] = None) -> None:
        """
        Set the OFDM parameters.

        Parameters
        ----------
        fft_size : int
            Size of the FFT and IFFT used by the OFDM class.
        cp_size : int
            Size of the cyclic prefix (in samples).
        num_used_subcarriers : int, optional
            Number of used subcarriers. Must be greater than or equal to 2
            and lower than or equal to fft_size. If not provided, fft_size
            will be used

        Raises
        ------
        ValueError
            If the any of the parameters are invalid.
        """
        if (cp_size < 0) or cp_size > fft_size:
            msg = ("cp_size must be nonnegative and cannot be greater "
                   "than fft_size")
            raise ValueError(msg)

        if num_used_subcarriers is None:
            num_used_subcarriers = fft_size

        if num_used_subcarriers > fft_size:
            msg = ("Number of used subcarriers cannot be greater than the "
                   "fft_size")
            raise ValueError(msg)

        if (num_used_subcarriers % 2 != 0) or (num_used_subcarriers < 2):
            msg = "Number of used subcarriers must be a multiple of 2"
            raise ValueError(msg)

        self.fft_size = fft_size
        self.cp_size = cp_size
        self.num_used_subcarriers = num_used_subcarriers

    def _calc_zeropad(self, input_data_size: int) -> Tuple[int, int]:
        """
        Calculates the number of zeros that must be added to the input data
        to make it a multiple of the OFDM size.

        The number of zeros that must be added to the input data is
        returned along with the number of OFDM symbols that will be
        generated.

        Parameters
        ----------
        input_data_size : int
            Size the the data that will be modulated by the OFDM object.

        Returns
        -------
        (zeropad, num_ofdm_symbols) : tuple[int,int]
            A tuple with zeropad and num_ofdm_symbols. Zeropad is the
            number of zeros added to the input data to make the total
            number of elements a multiple of the number of used
            subcarriers. Num_ofdm_symbols is the number of OFDM symbols
            required to transmit `input_data_size` symbols.
        """
        num_ofdm_symbols = (int(
            np.ceil(float(input_data_size) / self.num_used_subcarriers)))
        zeropad = (self.num_used_subcarriers * num_ofdm_symbols -
                   input_data_size)
        return zeropad, num_ofdm_symbols

    def _get_subcarrier_numbers(self) -> np.ndarray:
        """
        Get the indexes of all subcarriers, including the negative, the DC
        and the positive subcarriers.

        Note that these indexes are not suitable for indexing in
        python. They are the actual indexes of the subcarriers in an OFDM
        symbol. For instance, an OFDM symbol with 16 subcarriers will have
        indexes from -8 to 7. However, due to the way the fft is
        implemented in numpy the indexes here are actually from 0 to 7
        followed by -8 to -1.

        Returns
        -------
        np.ndarray
            Numbers of all subcarriers, including the negative, the DC and
            the positive subcarriers

        Examples
        --------
        >> ofdm_obj = OFDM(16, 4, 16)
        >> ofdm_obj._get_subcarrier_numbers()
        array([ 0,  1,  2,  3,  4,  5,  6,  7, \
                -8, -7, -6, -5, -4, -3, -2, -1])
        """
        indexes_regular_order = (np.arange(self.fft_size) - self.fft_size // 2)
        return np.fft.fftshift(indexes_regular_order)

    def _get_used_subcarrier_numbers(self) -> np.ndarray:
        """
        Get the subcarrier indexes of the actually used subcarriers.

        Note that these indexes are not suitable for indexing in
        python. They are the actual indexes of the subcarriers in an OFDM
        symbol. See the documentation of the _get_subcarrier_numbers
        function.

        Returns
        -------
        np.ndarray
            Number of the actually used subcarriers.

        Examples
        --------
        >> ofdm_obj = OFDM(16, 4, 10)
        >> ofdm_obj._get_used_subcarrier_numbers()
        array([ 1,  2,  3,  4,  5, -5, -4, -3, -2, -1])
        >> ofdm_obj = OFDM(16, 4, 14)
        >> ofdm_obj._get_used_subcarrier_numbers()
        array([ 1,  2,  3,  4,  5,  6,  7, -7, -6, -5, -4, -3, -2, -1])
        """
        if self.num_used_subcarriers == self.fft_size:
            return self._get_subcarrier_numbers()

        # Calculates half the number of subcarriers. This is only valid if
        # num_used_subcarriers is a multiple of 2.
        half_used_sc = self.num_used_subcarriers // 2

        first_half = np.r_[1:half_used_sc + 1]
        second_half = np.r_[-half_used_sc:0]
        indexes = np.hstack([first_half, second_half])
        return indexes

    def get_used_subcarrier_indexes(self) -> np.ndarray:
        """
        Get the subcarrier indexes of the subcarriers actually used in a
        way suitable for python indexing (going from 0 to fft_size-1).

        Returns
        -------
        indexes : np.ndarray
            Subcarrier indexes of the subcarriers actually used in a way
            suitable for python indexing.

        Notes
        -----
        This is the function actually used in the modulate function.

        Examples
        --------
        Consider the example below where we have 16 subcarriers and only
        10 subcarriers are used. The lower and higher subcarrier as well
        as the DC subcarrier will not be used. The index of the used
        subcarriers should go then from 11 to 15 (5 subcarriers),
        skip subcarrier 0, and then go from 1 to 5 (the other 5
        subcarriers).

        >>> ofdm_obj = OFDM(16, 4, 10)
        >>> ofdm_obj.get_used_subcarrier_indexes()
        array([11, 12, 13, 14, 15,  1,  2,  3,  4,  5])
        >>> ofdm_obj = OFDM(16,4,14)
        >>> ofdm_obj.get_used_subcarrier_indexes()
        array([ 9, 10, 11, 12, 13, 14, 15,  1,  2,  3,  4,  5,  6,  7])
        """
        numbers = self._get_used_subcarrier_numbers()
        half_used = self.num_used_subcarriers // 2

        indexes_proper = np.hstack(
            [self.fft_size + numbers[half_used:], numbers[0:half_used]])
        return indexes_proper

    def _prepare_input_signal(self, input_signal: np.ndarray) -> np.ndarray:
        """
        Prepare the input signal to be passed to the IFFT in the modulate
        function.

        The input signal must be prepared before it is passed to the IFFT
        in the OFDM modulate function.

        - First, zeros must be added so that the input signal size is
          multiple of the number of used subcarriers.
        - After that the input signal must be allocated to subcarriers in
          the center of the spectrum (except for the DC component). That
          is, zeros will be allocated to the lower and higher subcarriers
          such that only num_used_subcarriers are used from fft_size
          subcarriers.

        This preparation is performed by the _prepare_input_signal
        function.

        Parameters
        ----------
        input_signal : np.ndarray
            Input signal that must be modulated by the OFDM modulate
            function.

        Returns
        -------
        input_ifft : np.ndarray
            Signal suitable to be passed to the IFFT function to actually
            perform the OFDM modulation.

        See also
        --------
        _prepare_decoded_signal
        """
        # Number of data symbols passed to the modulate function
        num_symbs = input_signal.size
        # Calculates how many zeros need to be added as well as the number
        # of equivalent OFDM symbols.
        num_ofdm_symbols = self._calc_zeropad(num_symbs)[1]

        # Finally add the zeros to the input data
        input_signal = np.hstack([
            input_signal,
            np.zeros(self.num_used_subcarriers * num_ofdm_symbols - num_symbs)
        ])

        # Change the shape of the input data. Each row will be modulated as
        # one OFDM symbol.
        input_signal.shape = (num_ofdm_symbols, self.num_used_subcarriers)

        input_ifft = np.zeros([num_ofdm_symbols, self.fft_size], dtype=complex)
        input_ifft[:, self.get_used_subcarrier_indexes()] \
            = input_signal[:, :]

        return input_ifft

    def _prepare_decoded_signal(self,
                                decoded_signal: np.ndarray) -> np.ndarray:
        """
        Prepare the decoded signal that was processed by the FFT in the
        demodulate function.

        This is equivalent of reversing the indexing that was done by the
        _prepare_input_signal method.

        Parameters
        ----------
        decoded_signal : np.ndarray
            Signal that was decoded by the FFT in the OFDM demodulate
            method.

        Returns
        -------
        demodulated_samples : np.ndarray
            Demodulated samples of the symbols that were modulated by the
            OFDM object (for instance the PSK or M-QAM symbols passed to
            OFDM).

        Notes
        -----
        This method should be called AFTER the Cyclic Prefix was removed
        and the FFT was performed.

        Also, because the number of zeropad was not saved, then
        _prepare_decoded_signal has no way to remove them.

        See also
        --------
        _prepare_input_signal

        """
        return decoded_signal[:, self.get_used_subcarrier_indexes()].flatten()

    def _add_CP(self, input_data: np.ndarray) -> np.ndarray:
        """
        Add the Cyclic prefix to the input data.

        Parameters
        ----------
        input_data : np.ndarray
            OFDM modulated data (after the IFFT). This must be a 2D numpy
            array with shape (Number of OFDM symbols, IFFT size).

        Returns
        -------
        output : np.ndarray
            The `input_data` with the cyclic prefix added. The shape of the
            output is (Number of OFDM symbols, IFFT size + CP Size).

        """
        if self.cp_size != 0:
            output = np.hstack([input_data[:, -self.cp_size:], input_data])
        else:
            output = input_data
        return output

    def _remove_CP(self, received_data: np.ndarray) -> np.ndarray:
        """
        Remove the Cyclic prefix of the received data.

        Parameters
        ----------
        received_data : np.ndarray
            Data that must be demodulated by the OFDM object.

        Returns
        -------
        output : np.ndarray
            Received data without the Cyclic prefix.

        Notes
        -----
        The _remove_CP method will also change the shape so that it is
        suitable to be passed to the FFT function.

        """
        num_ofdm_symbols = (received_data.size //
                            (self.fft_size + self.cp_size))
        received_data.shape = (num_ofdm_symbols, self.fft_size + self.cp_size)
        received_data_no_CP = received_data[:, self.cp_size:]

        return received_data_no_CP

    def _calculate_power_scale(self) -> float:
        """
        Calculate the power scale that needs to be applied in the
        modulator and removed in the demodulate methods.

        The power is applied in the modulator method so that the total
        power of the OFDM samples is similar to the total power of the
        symbols modulated by OFDM.

        Note that this total power is shared among useful samples and the
        cyclic prefix in one OFDM symbol. Therefore, the larger the cyclic
        prefix size the lower is this power scale to account energy loss
        due to sending the cyclic prefix.

        Returns
        -------
        power_scale : float
            The calculated power scale. You should take the square root of
            this before multiplying by the samples.
        """
        power_scale = (float(self.fft_size) ** 2) / \
                      (float(self.num_used_subcarriers) + self.cp_size)
        return power_scale

    def modulate(self, input_signal: np.ndarray) -> np.ndarray:
        """
        Perform the OFDM modulation of the input_signal.

        Parameters
        ----------
        input_signal : np.ndarray
            Input signal that must be modulated by the OFDM modulate
            function.

        Returns
        -------
        output : np.ndarray
            An array with the samples of the modulated OFDM symbols.

        """
        # _prepare_input_signal will perform any zero padding needed as
        # well as deactivating the DC subcarrier and the guard subcarriers
        # when the number of used subcarriers is lower then the IFFT size.
        # Notice that the output of _prepare_input_signal will be a
        # bi-dimensional array, where each row has the input data for a
        # single OFDM symbol.
        input_ifft = self._prepare_input_signal(input_signal)

        # Now we calculate the ifft for the second axis. That is equivalent
        # to calculate the ifft separately for each row in the input_ifft
        # variable.
        output_ifft = (math.sqrt(self._calculate_power_scale()) *
                       np.fft.ifft(input_ifft, self.fft_size, 1))
        assert isinstance(output_ifft, np.ndarray)

        # Add the Cyclic prefix
        modulated_ofdm = self._add_CP(output_ifft)

        # Change the shape to one dimensional array and return that array
        return modulated_ofdm.flatten()

    def demodulate(self, received_signal: np.ndarray) -> np.ndarray:
        """
        Perform the OFDM demodulation of the received_signal.

        Parameters
        ----------
        received_signal : np.ndarray
            An array with the samples of the received OFDM symbols.

        Returns
        -------
        demodulated_data : np.ndarray
            Demodulated symbols.
        """
        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
        # - Remove the Cyclic Prefix -> the output will have a shape of
        # num_ofdm_symbols x fft_size
        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
        received_signal_no_CP = self._remove_CP(received_signal)

        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
        # - Call The FFT
        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
        # Now we calculate the FFT for the second axis. That is equivalent
        # to calculate the fft separately for each row in the
        # received_signal variable.
        output_fft = (np.fft.fft(received_signal_no_CP, self.fft_size, 1) /
                      math.sqrt(self._calculate_power_scale()))
        assert isinstance(output_fft, np.ndarray)

        # - Call the `_prepare_decoded_signal` method to get the data only
        # from the useful subcarriers
        decoded_symbols = self._prepare_decoded_signal(output_fft)

        # Return the decoded data
        return decoded_symbols


class OfdmOneTapEqualizer:
    """
    The OfdmOneTapEqualizer class performs the one-tap equalization often
    required in OFDM transmissions to compensate the effect of the channel
    at each subcarrier.

    Parameters
    ----------
    ofdm_obj : OFDM
        The OFDM object used to modulate/demodulate the data.
    """
    def __init__(self, ofdm_obj: OFDM):
        self._ofdm_obj = ofdm_obj

    def _equalize_data(self, data_reshaped: np.ndarray,
                       mean_freq_response: np.ndarray) -> np.ndarray:
        """
        Perform the one-tap equalization and return `data` after the
        channel compensation.

        Parameters
        ----------
        data_reshaped : np.ndarray
            The data to be equalized. If must be a 2D numpy array, where
            different rows correspond to different OFDM symbols and the
            different columns correspond to the USED
            subcarriers.
            Dimension: `num OFDM symbols x num Used subcarriers`
        mean_freq_response : np.ndarray
            The frequency response for each OFDM symbol.
            Dimension: `num OFDM symbols x FFT size`

        Returns
        -------
        np.ndarray
            The received `data` after the one-tap equalization to
            compensate the channel effect.
            Dimension: `num OFDM symbols x num Used subcarriers`
        """
        used_subcarriers_idx = self._ofdm_obj.get_used_subcarrier_indexes()

        equalized_ofdm_demodulated_data = \
            data_reshaped / mean_freq_response[:, used_subcarriers_idx]

        return equalized_ofdm_demodulated_data

    def equalize_data(
            self, data: np.ndarray,
            impulse_response: fading.TdlImpulseResponse) -> np.ndarray:
        """
        Perform the one-tap equalization and return `data` after the
        channel compensation.

        Parameters
        ----------
        data : np.ndarray
            The data to be equalized.
        impulse_response : fading.TdlImpulseResponse
            The impulse response of the channel.

        Returns
        -------
        np.ndarray
            The received `data` after the one-tap equalization to
            compensate the channel effect.
        """
        fft_size = self._ofdm_obj.fft_size
        num_used_subcarriers = self._ofdm_obj.num_used_subcarriers
        num_ofdm_symbols = data.size // num_used_subcarriers

        data_reshaped = np.reshape(data, (-1, num_used_subcarriers))

        freq_response = impulse_response.get_freq_response(fft_size)

        # Reshape and get the average frequency response for all samples in
        # each OFDM symbol
        freq_response = np.reshape(freq_response,
                                   (fft_size, num_ofdm_symbols, -1))
        mean_freq_response = np.mean(freq_response, axis=2)
        mean_freq_response = mean_freq_response.T

        equalized_ofdm_demodulated_data = self._equalize_data(
            data_reshaped, mean_freq_response)
        return equalized_ofdm_demodulated_data.flatten()



#!/usr/bin/env python
"""
Module implementing geometric shapes.

Each shape knows how to plot itself.
"""

try:
    # noinspection PyUnresolvedReferences
    from matplotlib import pyplot as plt
    # noinspection PyUnresolvedReferences
    from matplotlib import patches, path
    _MATPLOTLIB_AVAILABLE = True
except ImportError:  # pragma: no cover
    _MATPLOTLIB_AVAILABLE = False

import cmath
import math
from abc import ABCMeta, abstractmethod
from io import BytesIO
from typing import Any, Optional, TypeVar, cast

import numpy as np

__all__ = ['Coordinate', 'Shape', 'Hexagon', 'Rectangle', 'Circle']

ComplexOrArray = TypeVar("ComplexOrArray", np.ndarray, complex)


class Coordinate:
    """
    Base class for a coordinate in a 2D grid.

    A Coordinate object knows its location in the grid (represented as a
    complex number) and how to calculate the distance from it to another
    location.
    """
    def __init__(self, pos: complex):
        """
        Initializes the Coordinate object.

        Parameters
        ----------
        pos : complex
            Coordinate in the complex grid.
        """
        self._pos: complex = pos

    # xxxxxxxxxx pos property xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
    @property
    def pos(self) -> complex:
        """
        Get the coordinate position as a complex number.

        Returns
        -------
        complex
            The coordinate position (a complex number).
        """
        return self._pos

    @pos.setter
    def pos(self, value: complex) -> None:
        """
        Set the coordinate position.

        Parameters
        ----------
        value : complex
            The new coordinate position (a complex number).
        """
        self._pos = value

    # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

    def calc_dist(self, other: "Coordinate") -> float:
        """
        Calculates the distance to another coordinate.

        Parameters
        ----------
        other : Coordinate
            A different coordinate object.

        Returns
        -------
        dist : float
            Distance from self to the other coordinate.
        """
        dist: float = np.abs(self.pos - other.pos)
        return dist

    def move_by_relative_coordinate(self, rel_pos: complex) -> None:
        """
        Move from the current position to the relative coordinate.

        This is equivalent to moving to a new position given by the current
        position plus `coordinate`.

        Parameters
        ----------
        rel_pos : complex
            Relative coordinate
        """
        self.pos += rel_pos

    def move_by_relative_polar_coordinate(self, radius: float,
                                          angle: float) -> None:
        """
        Move from the current position to the relative coordinate.

        This is equivalent to moving to a new position given by the current
        position plus a the provided coordinate.

        Parameters
        ----------
        radius : float
            Distance of the movement in the direction given by `angle`.
        angle : float
            Angle (in radians) pointing the direction of the movement.
        """
        rel_pos = cmath.rect(radius, angle)
        self.move_by_relative_coordinate(rel_pos)

    def __repr__(self) -> str:  # pragma: no cover
        """
        Representation of a Coordinate object.

        Returns
        -------
        str
            The string representation of the object.
        """
        return "{0}({1})".format(self.__class__.__name__, self.pos)


class Shape(Coordinate):
    """
    Base class for all 2D shapes.

    Each subclass must implement the _get_vertex_positions method.

    Parameters
    ----------
    pos : complex
        Coordinate of the shape in the complex grid.
    radius : float
        Radius of the shape. It must be positive.
    rotation : float
        Rotation of the shape in degrees.
    """
    # The Shape class is an abstract class and all methods marked as
    # 'abstract' must be implemented in a subclass.
    __metaclass__ = ABCMeta

    def __init__(self, pos: complex, radius: float, rotation: float = 0):
        Coordinate.__init__(self, pos)

        self._radius = radius
        self._rotation = rotation

        # Properties for the plot representation
        self.fill_face_bool = False
        self.fill_color = 'r'
        self.fill_opacity = 0.1

        # Default figsize passed to matplotlib when the plot or the
        # _repr_some_format_ methods are called. Note that if you passed
        # the 'ax' argument in the plot method this will not be used.
        self.figsize = (8, 8)

    def __repr__(self) -> str:  # pragma: no cover
        """
        Representation of a Shape object.

        Returns
        -------
        str
            The string representation of the object.
        """
        return "{0}(pos={1},radius={2},rotation={3})".format(
            self.__class__.__name__, self.pos, self.radius, self.rotation)

    # xxxxx radius property xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
    @property
    def radius(self) -> float:
        """
        Get method for the radius property.

        Returns
        -------
        float
            The Shape radius.
        """
        return self._radius

    @radius.setter
    def radius(self, value: float) -> None:
        """
        Set method for the radius property.

        Parameters
        ----------
        value : float
            The new radius.
        """
        self._radius = value

    # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

    # xxxxx rotation property xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
    @property
    def rotation(self) -> float:
        """
        Get method for the rotation property.

        Returns
        -------
        float
            The shape rotation.
        """
        return self._rotation

    @rotation.setter
    def rotation(self, value: float) -> None:
        """
        Set method for the rotation property.

        Parameters
        ----------
        value : float
            The new shape rotation.
        """
        self._rotation = value

    # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

    @abstractmethod
    def _get_vertex_positions(self) -> np.ndarray:  # pragma: no cover
        """
        Calculates the vertex positions ignoring any rotation and
        considering that the shape is at the origin (rotation and
        translation will be added automatically later).

        Returns
        -------
        vertex_positions : np.ndarray
            The positions of the vertexes of the shape (as a 1D numpy
            array).

        Notes
        -----
        Not implemented. Must be implemented in a subclass and return a
        one-dimensional numpy array (complex dtype) with the vertex
        positions.
        """
        raise NotImplementedError(
            ('get_vertex_positions still needs to be implemented in the '
             '{0} class'.format(self.__class__.__name__)))

    # xxxxx vertex property xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
    @property
    def vertices_no_trans_no_rotation(self) -> np.ndarray:  # pragma: no cover
        """
        Get the shape vertexes without translation and rotation.

        Returns
        -------
        vertex_positions : np.ndarray
            The positions of the vertexes of the shape without any
            translation or rotation (as a 1D numpy array).
        """
        return self._get_vertex_positions()

    @property
    def vertices(self) -> np.ndarray:
        """
        Get method for the vertices property.

        Returns
        -------
        np.ndarray
            The shape vertexes.
        """
        vertex_positions: np.ndarray = self._get_vertex_positions()
        vertex_positions2: np.ndarray = self.pos + Shape.calc_rotated_pos(
            vertex_positions, self.rotation)
        return vertex_positions2

    # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

    def is_point_inside_shape(self, point: complex) -> bool:
        """
        Test is a point is inside the shape.

        Parameters
        ----------
        point
            A single complex number.

        Returns
        -------
        inside_or_not
            True if `point` is inside the shape, False otherwise.
        """
        mpl_path = path.Path(from_complex_array_to_real_matrix(self.vertices))

        # This code is used with Matplotlib version 1.2 or higher.
        return cast(bool, mpl_path.contains_point([point.real, point.imag]))

        # xxxxx Code for Matplotlib version 1.1 xxxxxxxxxxxxxxxxxxxxxxxxxxx
        # The code below was used for Matplotlib lower then version
        # 1.2. However, since Matplotlib version 1.2 the pnpoly function is
        # deprecated

        # # pnpoly returns 1 if point is inside the polygon and 0 otherwise
        # import matplotlib.nxutils as mnx
        # return mnx.pnpoly(point.real, point.imag,
        #                   from_complex_array_to_real_matrix(
        #                       self.vertices)) == 1

    # noinspection PyUnresolvedReferences
    def get_border_point(
            self,
            angle: float,
            ratio: Optional[float] = None) -> complex:  # pylint: disable=R0914
        """
        Calculates the coordinate of the point that intercepts the
        border of the shape if we go from the origin with a given angle
        (in degrees).

        Parameters
        ----------
        angle : float
            Angle in degrees.
        ratio : float
            The ratio from the cell center to the border where the desired
            point is located. This MUST be a value between 0 and 1.

        Returns
        -------
        point : complex
            A point in the line between the shape's center and the shape's
            border with the desired angle. If ratio is equal to one the
            point will be in the end of the line (touching the shape's
            border)
        """
        if ratio is None:
            ratio = 1.0

        angle_rad = np.pi * angle / 180.

        # Which point we get if we walk a distance of cell radius in the
        # desired angle direction?
        point = cast(complex, self.pos + self._radius * np.exp(angle_rad * 1j))

        # Calculates the distance of this point to all vertices and finds
        # the closest vertices
        dists = np.abs(self.vertices - point)
        # Get the two closest vertices from point
        closest_vertices = self.vertices[np.argsort(dists)[:2]]

        # The equation of a straight line is given by "y = ax + b". We have
        # two points in this line (the two closest vertices) and we can use
        # them to find 'a' and 'b'. First let's find the different of these
        # two closest vertexes
        diff = closest_vertices[0] - closest_vertices[1]

        # xxxxx Special case for a vertical line xxxxxxxxxxxxxxxxxxxxxxxxxx
        # noinspection PyTypeChecker
        if np.allclose(diff.real, 0.0, atol=1e-15):
            # If the the real part of diff is equal to zero, that means
            # that the straight line is actually a vertical
            # line. Therefore, all we need to do to get the border point is
            # to start from the shape's center and go with the desired
            # angle until the value in the 'x' axis is equivalent to
            # closest_vertices[0].real.
            adjacent_side = closest_vertices[0].real - self.pos.real
            side = np.tan(angle_rad) * adjacent_side
            point = self.pos + adjacent_side + 1j * side
            # Now all that is left to do is apply the ratio, which only
            # means that the returned point is a linear combination between
            # the shape's central position and the point at the border of
            # the shape

            return (1 - ratio) * self.pos + ratio * point
        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

        # Calculates the 'a' and 'b' in the line equation "y=ax+b"
        a = diff.imag / diff.real
        b = closest_vertices[1].imag - a * closest_vertices[1].real

        # Note that is we start from self.pos and walk in the direction
        # pointed by the angle by "some step" we should reach the line
        # where the two closest vertexes are. If we can find this "step"
        # then we will get our desired point.
        # That is, for the step "z" we have
        #    self.pos + np.exp(1j * angle_rad) * z = complex(x, a * x + b)
        # Which we can write as the system of equations
        #    self.pos.real + np.exp(1j * angle).real * z = x
        #    self.pos.imag + np.exp(1j * angle).imag * z = a * x + b
        # Lets create some aliases for the constants so that
        #     A + B * z = x
        #     C + D * z = a * x + b
        A = self.pos.real
        B = np.exp(1j * angle_rad).real
        C = self.pos.imag
        D = np.exp(1j * angle_rad).imag
        # Through some algebraic manipulation the correct step "z" is given
        # by
        z = (A * a + b - C) / (D - (a * B))

        # Now we can finally find the desired point at the border of the
        # shape
        point = self.pos + np.exp(1j * angle_rad) * z

        # Now all that is left to do is apply the ratio, which only means
        # that the returned point is a linear combination between the
        # shape's central position and the point at the border of the shape
        return (1 - ratio) * self.pos + ratio * point

    # noinspection PyShadowingNames,PyShadowingNames
    def plot(self, ax: Any = None) -> None:  # pragma: no cover
        """
        Plot the shape using the matplotlib library.

        Parameters
        ----------
        ax : A matplotlib ax, optional
            The ax where the shape will be plotted. If not provided, a new
            figure (and ax) will be created.

        Notes
        -----
        If an axes 'ax' is specified, then the shape is added to that
        axes. Otherwise a new figure and axes are created and the shape is
        plotted to that.
        """
        stand_alone_plot = False

        if ax is None:
            # This is a stand alone plot. Lets create a new axes.
            _, ax = plt.subplots(figsize=self.figsize)
            stand_alone_plot = True

        if self.fill_face_bool:
            # Matplotlib does not seem to allow us to set a different alpha
            # value for the face and the edges. Therefore, we will need to
            # plot twice to get that effect.
            polygon_face = patches.Polygon(
                from_complex_array_to_real_matrix(self.vertices),
                True,
                facecolor=self.fill_color,
                edgecolor='none',  # No edges
                alpha=self.fill_opacity)
            ax.add_patch(polygon_face)

        polygon_edges = patches.Polygon(
            from_complex_array_to_real_matrix(self.vertices),
            True,
            facecolor='none',  # No face
            edgecolor="black",
            alpha=1)

        ax.add_patch(polygon_edges)

        if stand_alone_plot is True:
            ax.plot()
            plt.show()

    # noinspection PyShadowingNames
    def _repr_some_format_(
            self,
            extension: str = 'png',
            axis_option: str = 'equal') -> Any:  # pragma: nocover
        """
        Return the representation of the shape in the desired format.

        Parameters
        ----------
        extension : str
            The extension of the desired format. This should be something
            that the savefig method in a matplotlib figure can understand,
            such as 'png', 'svg', stc.
        axis_option : str
            Option to be given to the ax.axis function.

        Returns
        -------
        output
        """
        plt.ioff()  # turn off interactive mode
        fig = plt.figure(figsize=self.figsize)
        ax = fig.add_subplot(111)
        ax.set_axis_off()

        output = BytesIO()
        self.plot(ax)
        ax.axis(axis_option)
        fig.savefig(output, format=extension)
        output.seek(0)
        plt.close(fig)
        plt.ion()  # turn on interactive mode

        return output.getvalue()

    def _repr_png_(self) -> Any:  # pragma: no cover
        """
        Return the PNG representation of the shape.
        """
        return self._repr_some_format_('png')

    def _repr_svg_(self) -> Any:  # pragma: no cover
        """
        Return the SVG representation of the shape.
        """
        return self._repr_some_format_('svg')

    @staticmethod
    def calc_rotated_pos(cur_pos: ComplexOrArray,
                         angle: float) -> ComplexOrArray:
        """
        Rotate the complex numbers in the `cur_pos` array by `angle` (in
        degrees) around the origin.

        Parameters
        ----------
        cur_pos : complex | np.ndarray
            The complex number(s) to be rotated.
        angle: float
            Angle in degrees to rotate the positions.

        Returns
        -------
        rotated_pos : complex | np.ndarray
            The rotate complex number(s).
        """
        angle_rad = angle * np.pi / 180.
        return cur_pos * np.exp(1j * angle_rad)  # type: ignore


class Hexagon(Shape):
    """
    Hexagon shape class.

    Besides the `pos`, `radius` and `rotation` properties from the Shape
    base class, the Hexagon also has a height property (read-only) from the
    base of the Hexagon to its center.

    Parameters
    ----------
    pos : complex
        Coordinate of the shape in the complex grid.
    radius : float
        Radius of the hexagon. It must be a positive number.
    rotation : float
        Rotation of the hexagon in degrees.
    """
    def __init__(self, pos: complex, radius: float, rotation: float = 0):
        Shape.__init__(self, pos, radius, rotation)

    @property
    def height(self) -> float:
        """
        Get method for the height property.

        Returns
        -------
        float
            The height of the Hexagon.
        """
        return self._radius * math.sqrt(3.) / 2.0

    def _get_vertex_positions(self) -> np.ndarray:
        """
        Calculates the vertex positions ignoring any rotation and
        considering that the hexagon is at the origin (rotation and
        translation will be added automatically later).

        Returns
        -------
        vertex_positions : np.ndarray
            The positions of the vertexes of the shape.
        """
        vertex_positions: np.ndarray = np.zeros(6, dtype=complex)
        vertex_positions[0] = complex(-self._radius / 2., -self.height)
        # noinspection PyTypeChecker
        angles = np.linspace(0, 240, 5) * np.pi / 180.

        for k in range(5):
            # noinspection PyUnresolvedReferences
            vertex_positions[k + 1] = (vertex_positions[k] +
                                       self._radius * np.exp(angles[k] * 1j))

        return vertex_positions


class Rectangle(Shape):
    """
    Rectangle shape class.

    The rectangle is initialized from two coordinates as well as from
    the rotation.

    Parameters
    ----------
    first : complex
        First coordinate (without rotation).
    second : complex
        Second coordinate (without rotation).
    rotation : float
        Rotation of the rectangle in degrees.
    """
    def __init__(self, first: complex, second: complex, rotation: float = 0):
        central_pos = (first + second) / 2
        radius = np.abs(second - central_pos)
        Shape.__init__(self, central_pos, radius, rotation)
        self._lower_coord = complex(min(first.real, second.real),
                                    min(first.imag, second.imag))
        self._upper_coord = complex(max(first.real, second.real),
                                    max(first.imag, second.imag))

    def __repr__(self) -> str:  # pragma: no cover
        """
        Representation of a Rectangle object.

        Returns
        -------
        str
            The string representation of the Rectangle object.
        """
        return "{0}(A={1},B={2},rotation={3})".format(self.__class__.__name__,
                                                      self._lower_coord,
                                                      self._upper_coord,
                                                      self.rotation)

    def _get_vertex_positions(self) -> np.ndarray:
        """
        Calculates the vertex positions ignoring any rotation and
        considering that the rectangle is at the origin (rotation and
        translation will be added automatically later).

        Returns
        -------
        vertex_positions : np.ndarray
            The positions of the vertexes of the shape.
        """
        vertex_positions: np.ndarray = np.zeros(4, dtype=complex)
        A = self._lower_coord - self.pos
        B = self._upper_coord - self.pos
        vertex_positions[0] = A
        vertex_positions[1] = complex(B.real, A.imag)
        vertex_positions[2] = B
        vertex_positions[3] = complex(A.real, B.imag)
        return vertex_positions

    def _repr_some_format_(
            self,
            extension: str = 'png',
            axis_option: str = 'tight') -> Any:  # pragma: no cover
        """
        Return the representation of the shape in the desired format.

        Parameters
        ----------
        extension : str
            The extension of the desired format. This should be something
            that the savefig method in a matplotlib figure can understand,
            such as 'png', 'svg', stc.
        axis_option : str
            Option to be given to the ax.axis function.

        Notes
        -----
        We only subclass the _repr_some_format_ method from the Shape class
        here so that we can change the axis_option to 'tight' (default in
        the Shape class is 'equal').
        """
        return Shape._repr_some_format_(self,
                                        extension=extension,
                                        axis_option=axis_option)

    def is_point_inside_shape(self, point: complex) -> bool:
        """
        Test is a point is inside the rectangle

        Parameters
        ----------
        point : complex
            A single complex number.

        Returns
        -------
        bool
            True if `point` is inside the rectangle, False otherwise.
        """
        min_x = min(self._lower_coord.real, self._upper_coord.real)
        max_x = max(self._lower_coord.real, self._upper_coord.real)
        min_y = min(self._lower_coord.imag, self._upper_coord.imag)
        max_y = max(self._lower_coord.imag, self._upper_coord.imag)

        point_x = point.real
        point_y = point.imag
        if point_x < min_x:
            return False
        if point_x > max_x:
            return False
        if point_y < min_y:
            return False
        if point_y > max_y:
            return False
        return True


class Circle(Shape):
    """
    Circle shape class.

    A circle is initialized only from a coordinate and a radius.

    Parameters
    ----------
    pos : complex
        Coordinate of the center of the circle.
    radius : float
        Circle's radius.
    """
    def __init__(self, pos: complex, radius: float):
        Shape.__init__(self, pos, radius)

    def _get_vertex_positions(self) -> np.ndarray:
        """
        Calculates the vertex positions considering that the circle is
        at the origin (translation will be added automatically later).

        Returns
        -------
        vertex_positions : np.ndarray
            The positions of the vertexes of the shape.

        Notes
        -----
        It does not make much sense to get the vertexes of a circle, since
        a circle 'has' infinite vertexes. However, for consistence with the
        Shape's class interface the _get_vertex_positions is implemented
        such that it returns a subset of the circle vertexes. The number of
        returned vertexes was arbitrarily chosen as 12.
        """
        num_vertexes = 12
        angles: np.ndarray = np.linspace(
            0, (num_vertexes - 1.) / num_vertexes * 2 * np.pi, num_vertexes)

        vertex_positions: np.ndarray = self._radius * np.exp(1j * angles)
        return vertex_positions

    def get_border_point(self,
                         angle: float,
                         ratio: Optional[float] = None) -> complex:
        """
        Calculates the coordinate of the point that intercepts the
        border of the circle if we go from the origin with a given angle
        (in degrees).

        Parameters
        ----------
        angle : float
            Angle in degrees
        ratio : float
            The ratio from the cell center to the border where the desired
            point is located. It must be a value between 0 and 1.

        Returns
        -------
        point : complex
            A point in the line between the circle's center and the
            circle's border with the desired angle. If ratio is equal to
            one the point will be in the end of the line (touching the
            circle's border)
        """
        if ratio is None:
            ratio = 1.0

        angle_rad = np.pi * angle / 180.
        return cast(complex,
                    self.pos + np.exp(1j * angle_rad) * self.radius * ratio)

    def is_point_inside_shape(self, point: complex) -> bool:
        """
        Test is a point is inside the circle

        Parameters
        ----------
        point
            A single complex number.

        Returns
        -------
        inside_or_not
            True if `point` is inside the circle, False otherwise.
        """
        return cast(bool, np.abs(self.pos - point) < self.radius)

    # noinspection PyShadowingNames,PyShadowingNames
    def plot(self, ax: Any = None) -> None:  # pragma: no cover
        """
        Plot the circle using the Matplotlib library.

        Parameters
        ----------
        ax : A matplotlib axis, optional
            The axis where the shape will be plotted. If not provided,
            a new figure (and axis) will be created.

        Notes
        -----
        If an axes 'ax' is specified, then the shape is added to that
        axes. Otherwise a new figure and axes are created and the shape is
        plotted to that.
        """
        stand_alone_plot = False

        if ax is None:
            # This is a stand alone plot. Lets create a new axes.
            ax = plt.axes()
            stand_alone_plot = True

        if self.fill_face_bool:
            # Matplotlib does not seem to allow us to set a different alpha
            # value for the face and the edges. Therefore, we will need to
            # plot twice to get that effect.
            circle_face = patches.Circle(
                [self.pos.real, self.pos.imag],
                self.radius,
                facecolor=self.fill_color,
                edgecolor='none',  # No edges
                alpha=self.fill_opacity)
            ax.add_patch(circle_face)

        circle_edges = patches.Circle(
            [self.pos.real, self.pos.imag],
            self.radius,
            facecolor='none',  # No face
            alpha=1)
        ax.add_patch(circle_edges)

        if stand_alone_plot is True:
            ax.plot()
            plt.show()


# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
def from_complex_array_to_real_matrix(a: np.ndarray) -> np.ndarray:
    """
    Convert an array of complex numbers to a matrix of real numbers.

    Parameters
    ----------
    a : np.ndarray
        A numpy array of complex numbers with N elements.

    Returns
    -------
    np.ndarray
        The converted array with dimension N x 2. That is, a 2D numpy
        array.

    Notes
    -----
    We use complex number to represent coordinates, where the real part is
    the 'x' coordinate and the imaginary part is the 'y' coordinate pf a
    point.

    However, matplotlib methods need the coordinates to be separated. For
    instance, a vector of complex coordinates must be converted to a matrix
    with two columns, where the two columns are the 'x' and 'y' coordinates
    and each row corresponds to a point.

    The method from_complex_array_to_real_matrix does exactly this
    conversion.

    Examples
    --------
    >>> a = np.array([1+2j, 3-4j, 5+6j])
    >>> print(from_complex_array_to_real_matrix(a))
    [[ 1.  2.]
     [ 3. -4.]
     [ 5.  6.]]
    """
    num_elem = np.size(a)
    a.dtype = a.real.dtype
    a = np.reshape(a, (num_elem, 2))
    return a


# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
if __name__ == '__main__1':  # pragma: no cover
    ax = plt.axes()
    h = Hexagon(2 + 3j, 2, 30)

    # print "Border Point is: {0}".format(point)
    h.plot(ax)

    POINT1 = h.get_border_point(90, 0.9)
    ax.plot(POINT1.real, POINT1.imag, 'ro')

    POINT2 = h.get_border_point(10, 0.9)
    ax.plot(POINT2.real, POINT2.imag, 'go')

    POINT3 = h.get_border_point(30, 0.9)
    ax.plot(POINT3.real, POINT3.imag, 'bo')

    POINT4 = h.get_border_point(107, 1)
    ax.plot(POINT4.real, POINT4.imag, 'bo')

    ax.plot(h.pos.real, h.pos.imag, 'ro')

    # print h.vertices
    ax.axis('equal')
    plt.show()

if __name__ == '__main__':  # pragma: no cover
    ax = plt.axes()

    h = Hexagon(0, 1)
    h.rotation = 30
    h.fill_face_bool = True

    plt.hold(True)
    r = Rectangle(0, 2 + 1j, 0)
    r.rotation = 15
    r.fill_face_bool = True
    r.fill_color = 'b'

    c = Circle(0.2 + 0.4j, 1)
    c.fill_face_bool = True
    c.fill_color = 'g'

    r.plot(ax)
    h.plot(ax)
    c.plot(ax)

    ax.plot()
    plt.axis('equal')
    plt.show()

if __name__ == '__main__1':  # pragma: no cover
    ax = plt.axes()
    c = Circle(2 + 3j, 2)

    # print "Border Point is: {0}".format(point)
    c.plot(ax)

    for v in c.vertices:
        ax.plot(v.real, v.imag, 'bo')

    ax.plot(c.pos.real, c.pos.imag, 'ro')

    # print c.vertices
    ax.axis('equal')
    plt.show()



#!/usr/bin/env python

# http://www.doughellmann.com/PyMOTW/struct/
# import struct
# import binascii
"""
Module with class for some fundamental modulators, such as PSK and M-QAM.

All fundamental modulators inherit from the `Modulator` class and should
call the self.setConstellation method in their __init__ method, as well
as implement the calcTheoreticalSER and calcTheoreticalBER methods. """

try:
    # noinspection PyUnresolvedReferences
    import matplotlib.pyplot as plt
    _MATPLOTLIB_AVAILABLE = True
except ImportError:  # pragma: no cover
    _MATPLOTLIB_AVAILABLE = False

import math
from typing import Optional, TypeVar, Union

import numpy as np

from pyphysim.util.conversion import binary2gray, dB2Linear, gray2binary
from pyphysim.util.misc import level2bits, qfunc

PI = np.pi

NumberOrArray = TypeVar("NumberOrArray", np.ndarray, float)

__all__ = ['Modulator', 'PSK', 'QPSK', 'BPSK', 'QAM']


# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxx Modulator Class xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
class Modulator:
    """
    Base class for digital modulators.

    The derived classes need to at least call setConstellation to set the
    constellation in their constructors as well as implement
    calcTheoreticalSER and calcTheoreticalBER.

    Examples
    --------
    >>> np.set_printoptions(linewidth=70)
    >>> constellation = np.array([1 + 1j, - 1 + 1j, - 1 - 1j, 1 - 1j])
    >>> m=Modulator()
    >>> m.setConstellation(constellation)
    >>> m.symbols
    array([ 1.+1.j, -1.+1.j, -1.-1.j,  1.-1.j])
    >>> m.M
    4
    >>> m.K
    2.0
    >>> m
    4-Modulator object
    >>> m.modulate(np.array([0, 0, 3, 3, 1, 3, 3, 3, 2, 2]))
    array([ 1.+1.j,  1.+1.j,  1.-1.j,  1.-1.j, -1.+1.j,  1.-1.j,  1.-1.j,
            1.-1.j, -1.-1.j, -1.-1.j])

    >>> m.demodulate(np.array([ 1. + 1.j, 1. + 1.j, 1. - 1.j, 1. - 1.j, \
                                - 1. + 1.j, 1. - 1.j, 1. - 1.j, 1. - 1.j, \
                                - 1. - 1.j, - 1. - 1.j]))
    array([0, 0, 3, 3, 1, 3, 3, 3, 2, 2])
    """
    def __init__(self) -> None:
        """Initializes the Modulator object.
        """
        # This should be set in a subclass of the Modulator Class by
        # calling the setConstellation method..
        self._M: int = 0  # Constellation size (modulation cardinality)
        # Number of bits represented by each symbol in the constellation
        self._K: int = 0
        self.symbols: np.ndarray = np.array([])

    @property
    def name(self) -> str:
        """
        Get method for the 'name' property.

        Returns
        -------
        str
            The name of the modulator.
        """
        return "{0:d}-{1:s}".format(self._M, self.__class__.__name__)

    @property
    def M(self) -> int:
        """
        Get method for the M property.

        The `M` property corresponds to the number of symbols in the
        constellation.

        See also
        --------
        K
        """
        return self._M

    @property
    def K(self) -> int:
        """
        Get method for the K property.

        The `K` property corresponds to the number of bits represented by
        each symbol in the constellation. It is equal to log2(M), where `M`
        is the constellation size.

        See also
        --------
        M
        """
        return self._K

    def __repr__(self) -> str:  # pragma: no cover
        """
        Get the string representation of the object.

        Returns
        -------
        str
            String representation of the object.
        """
        return "{0} object".format(self.name)

    def setConstellation(self, symbols: np.ndarray) -> None:
        """
        Set the constellation of the modulator.

        This function should be called in the constructor of the derived
        classes.

        Parameters
        ----------
        symbols : np.ndarray
            A an numpy array with the symbol table.
        """
        M = symbols.size
        self._M = M
        self._K = np.log2(M)
        self.symbols = symbols

    def plotConstellation(self) -> None:  # pragma: no cover
        """Plot the constellation (in a scatter plot).
        """
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)  # one row, one column, first plot
        # circlePatch = patches.Circle((0,0),1)
        # circlePatch.set_fill(False)
        # ax.add_patch(circlePatch)
        ax.scatter(self.symbols.real, self.symbols.imag)

        ax.axis('equal')
        ax.grid()

        formatString = "{0:0=" + str(level2bits(self._M)) + "b} ({0})"

        index = 0
        for symbol in self.symbols:
            ax.text(
                symbol.real,  # Coordinate X
                symbol.imag + 0.03,  # Coordinate Y
                formatString.format(index, format_spec="0"),  # Text
                verticalalignment='bottom',  # From now on, text properties
                horizontalalignment='center')
            index += 1

        plt.show()

    def modulate(self, inputData: Union[int, np.ndarray]) -> np.ndarray:
        """
        Modulate the input data (decimal data).

        Parameters
        ----------
        inputData : np.ndarray | int
            Data to be modulated.

        Returns
        -------
        modulated_data : np.ndarray
            The modulated data

        Raises
        ------
        ValueError
            If inputData has any invalid value such as values greater
            than self._M - 1. Note that inputData should not have
            negative values but no check is done for this.
        """
        try:
            return self.symbols[inputData]
        except IndexError:
            raise ValueError("Input data must be between 0 and 2^M")

    def demodulate(self, receivedData: np.ndarray) -> np.ndarray:
        """
        Demodulate the data.

        Parameters
        ----------
        receivedData : np.ndarray
            Data to be demodulated.

        Returns
        -------
        demodulated_data : np.ndarray
            The demodulated data.
        """
        # xxxxxxxxxx First Try xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
        # def getClosestSymbol(symb):
        #     closestSymbolIndex = np.abs(self.symbols - symb).argmin()
        #     return closestSymbolIndex
        # getClosestSymbol = np.frompyfunc(getClosestSymbol, 1, 1)
        # return getClosestSymbol(receivedData).astype(int)
        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

        # ### Second Try xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
        # # This versions is a little faster then the first version
        # shape = receivedData.shape
        # num_symbols = receivedData.size
        # output = np.empty(num_symbols, dtype=int)
        # reshaped_received_data = receivedData.flatten()

        # for ii in range(num_symbols):
        #     output[ii] = np.abs(
        #                  self.symbols
        #                  - reshaped_received_data[ii]).argmin()
        # output.shape = shape
        # return output
        # # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

        # xxxxxxxxxx Third Try xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
        # This version uses more memory because of the numpy broadcasting,
        # but it is much faster.
        shape = receivedData.shape
        reshaped_received_data = receivedData.flatten()

        constellation = np.reshape(self.symbols, [self.symbols.size, 1])
        output = np.abs(constellation - reshaped_received_data).argmin(axis=0)
        output.shape = shape

        return output
        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

    def calcTheoreticalSER(
            self, SNR: NumberOrArray) -> NumberOrArray:  # pragma: no cover
        """
        Calculates the theoretical symbol error rate.

        Parameters
        ----------
        SNR : float | np.ndarray
            Signal-to-noise-value (in dB).

        Returns
        -------
        SER : float | np.ndarray
            The theoretical symbol error rate.

        See also
        --------
        calcTheoreticalBER,
        calcTheoreticalPER

        Notes
        -----
        This function should be implemented in the derived classes
        """
        raise NotImplementedError("calcTheoreticalSER: Not implemented")

    def calcTheoreticalBER(
            self, SNR: NumberOrArray) -> NumberOrArray:  # pragma: no cover
        """
        Calculates the theoretical bit error rate.

        Parameters
        ----------
        SNR : float | np.ndarray
            Signal-to-noise-value (in dB).

        Returns
        -------
        BER : float | np.ndarray
            The theoretical bit error rate.

        See also
        --------
        calcTheoreticalSER,
        calcTheoreticalPER

        Notes
        -----
        This function should be implemented in the derived classes
        """
        raise NotImplementedError("calcTheoreticalBER: Not implemented")

    def calcTheoreticalPER(self, SNR: NumberOrArray,
                           packet_length: int) -> NumberOrArray:
        """
        Calculates the theoretical package error rate.

        A package is a group of bits, where if a single bit is in error
        then the whole package is considered to be in error.

        The package error rate (PER) is a direct mapping of the bit error
        rate (BER), such that

        .. math::
           PER = 1 - (1 - BER)^{L}

        where :math:`L` is the package_length.

        Parameters
        ----------
        SNR : float | np.ndarray
            Signal-to-noise-value (in dB).
        packet_length : int
            The package length. That is, the number of bits in each
            package.

        Returns
        -------
        PER : float | np.ndarray
            The theoretical package error rate.

        See also
        --------
        calcTheoreticalBER,
        calcTheoreticalSER
        calcTheoreticalSpectralEfficiency
        """
        BER = self.calcTheoreticalBER(SNR)
        PER = 1 - ((1 - BER)**packet_length)
        return PER

    def calcTheoreticalSpectralEfficiency(
            self,
            SNR: NumberOrArray,
            packet_length: Optional[int] = None) -> NumberOrArray:
        """
        Calculates the theoretical spectral efficiency.

        If there was no error in the transmission, the spectral efficiency
        would be equal to the `K` property, that is, equal to the number of
        bits represented by each symbol in the constellation. However, due
        to bit errors the effective spectral efficiency will be lower.

        The calcTheoreticalSpectralEfficiency method calculates the
        effective spectral efficiency from the `K` property and the package
        error rate (PER) for the given SNR and packet_length 'L', such that

        .. math::
           se = K * (1 - PER)

        Parameters
        ----------
        SNR : float | np.ndarray
            Signal-to-noise-value (in dB).
        packet_length : int, optional
            The package length. That is, the number of bits in each
            package.

        Returns
        -------
        se : float | np.ndarray
            The theoretical spectral efficiency.

        See also
        --------
        calcTheoreticalBER,
        calcTheoreticalPER,
        K
        """
        if packet_length is None:
            se = self.K * (1 - self.calcTheoreticalBER(SNR))
        else:
            se = self.K * (1 - self.calcTheoreticalPER(SNR, packet_length))
        return se


# xxxxx End of Modulator Class xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx


# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxx PSK Class xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
class PSK(Modulator):
    """PSK Class
    """
    def __init__(self, M: int, phaseOffset: float = 0) -> None:
        """Initializes the PSK object.

        Parameters
        ----------
        M : int
            The modulation cardinality
        phaseOffset: float, optional
            A phase offset (in radians) to be applied to the PSK
            constellation.

        """
        Modulator.__init__(self)
        # Check if M is a power of 2
        assert 2**math.log(M, 2) == M

        # Generates the constellation
        symbols = self._createConstellation(M, phaseOffset)

        # Change to Gray mapping
        # noinspection PyUnresolvedReferences
        symbols = symbols[gray2binary(np.arange(0, M))]

        self.setConstellation(symbols)

    # noinspection PyUnresolvedReferences
    @staticmethod
    def _createConstellation(M: int, phaseOffset: float) -> np.ndarray:
        """Generates the Constellation for the PSK modulation scheme.

        Parameters
        ----------
        M : int
            The modulation cardinality
        phaseOffset: float
            A phase offset (in radians) to be applied to the PSK
            constellation.

        Returns
        -------
        symbols : np.ndarray
            The PSK constellation with the desired cardinality and phase
            offset.

        """
        phases = 2. * PI / M * np.arange(0, M) + phaseOffset
        realPart = np.cos(phases)
        imagPart = np.sin(phases)

        # Any number inferior to 1e-15 will be considered as 0
        realPart[abs(realPart) < 1e-15] = 0
        imagPart[abs(imagPart) < 1e-15] = 0
        return realPart + 1j * imagPart

    def setPhaseOffset(self, phaseOffset: float) -> None:
        """Set a new phase offset for the constellation

        Parameters
        ----------
        phaseOffset: float
            A phase offset (in radians) to be applied to the PSK
            constellation.
        """
        self.setConstellation(self._createConstellation(self._M, phaseOffset))

    # noinspection PyPep8
    def calcTheoreticalSER(self, SNR: NumberOrArray) -> NumberOrArray:
        """Calculates the theoretical (approximation for high M and high
        SNR) symbol error rate for the M-PSK scheme.

        Parameters
        ----------
        SNR : float | np.ndarray
            Signal-to-noise-value (in dB).

        Returns
        -------
        SER : float | np.ndarray
            The theoretical symbol error rate.
        """
        snr = dB2Linear(SNR)

        # $P_s \approx 2Q\left(\sqrt{2\gamma_s}\sin\frac{\pi}{M}\right)$
        # Alternative formula (same result)
        # $P_s = erfc \left ( \sqrt{\gamma_s} \sin(\frac{\pi}{M})  \right )$
        ser = 2. * qfunc(np.sqrt(2. * snr) * math.sin(PI / self._M))
        return ser

    def calcTheoreticalBER(self, SNR: NumberOrArray) -> NumberOrArray:
        """Calculates the theoretical (approximation) bit error rate for
        the M-PSK scheme using Gray coding.

        Parameters
        ----------
        SNR : float | np.ndarray
            Signal-to-noise-value (in dB).

        Returns
        -------
        BER : float | np.ndarray
            The theoretical bit error rate.
        """
        # $P_b = \frac{1}{k}P_s$
        # Number of bits per symbol
        k = level2bits(self._M)
        return 1.0 / k * self.calcTheoreticalSER(SNR)


# xxxxx End of PSK Class xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx


# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxx QPSK Class xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
class QPSK(PSK):  # pragma: no cover
    """QPSK Class
    """
    def __init__(self) -> None:
        PSK.__init__(self, 4, PI / 4.)

    def __repr__(self) -> str:  # pragma: no cover
        """
        Get the string representation of the object.

        Returns
        -------
        str
            String representation of the object.
        """
        return "QPSK object"


# xxxxx End of QPSK Class xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx


# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxx BPSK Class xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
class BPSK(Modulator):
    """BPSK Class
    """
    def __init__(self) -> None:
        Modulator.__init__(self)
        # The number "1" will be mapped to "-1" and the number "0" will be
        # mapped to "1"
        self.setConstellation(np.array([1, -1]))

    @property
    def name(self) -> str:
        """
        Get the name property.

        Returns
        -------
        str
            The name of the modulator.
        """
        return "{0:s}".format(self.__class__.__name__)

    def __repr__(self) -> str:  # pragma: no cover
        """
        Get the string representation of the object.

        Returns
        -------
        str
            String representation of the object.
        """
        return "BPSK object"

    def calcTheoreticalSER(self, SNR: NumberOrArray) -> NumberOrArray:
        """
        Calculates the theoretical (approximation) symbol error rate for
        the BPSK scheme.

        Parameters
        ----------
        SNR : float | np.ndarray
            Signal-to-noise-value (in dB).

        Returns
        -------
        SER : float | np.ndarray
            The theoretical symbol error rate.
        """
        snr = dB2Linear(SNR)
        # $P_b = Q\left(\sqrt{\frac{2E_b}{N_0}}\right)$
        # Alternative formula (same result)
        # $P_b = \frac{1}{2}erfc \left ( \sqrt{\frac{E_b}{N_0}} \right )$
        ser = qfunc(np.sqrt(2 * snr))
        return ser

    def calcTheoreticalBER(self, SNR: NumberOrArray) -> NumberOrArray:
        """
        Calculates the theoretical (approximation) bit error rate for the
        BPSK scheme.

        Parameters
        ----------
        SNR : float | np.ndarray
            Signal-to-noise-value (in dB).

        Returns
        -------
        BER : float | np.ndarray
            The theoretical bit error rate.
        """
        return self.calcTheoreticalSER(SNR)

    def modulate(self, inputData: np.ndarray) -> np.ndarray:
        """
        Modulate the input data (decimal data).

        Parameters
        ----------
        inputData : np.ndarray
            Data to be modulated.

        Returns
        -------
        modulated_data : np.ndarray
            The modulated data

        Raises
        ------
        ValueError
            If inputData has any invalid value such as values greater
            than self._M - 1. Note that inputData should not have
            negative values but no check is done for this.

        """
        # noinspection PyTypeChecker
        if np.any(inputData > 1):
            raise ValueError("Input data can only contains '0's and '1's")
        return 1 - 2 * inputData

    def demodulate(self, receivedData: np.ndarray) -> np.ndarray:
        """
        Demodulate the data.

        Parameters
        ----------
        receivedData : np.ndarray
            Data to be demodulated.

        Returns
        -------
        demodulated_data : np.ndarray
            The demodulated data.
        """
        # noinspection PyUnresolvedReferences
        return (receivedData < 0).astype(int)


# xxxxx End of BPSK Class xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx


# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxx QAM Class xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
class QAM(Modulator):
    """QAM Class
    """
    def __init__(self, M: int) -> None:
        """Initializes the QAM object.

        Parameters
        ----------
        M : int
            The modulation cardinality

        Raises
        ------
        ValueError
            If M is not a square power of 2.
        """
        Modulator.__init__(self)

        # Check if M is an even power of 2
        power = math.log(M, 2)
        if (power % 2 != 0) or (2**power != M):
            raise ValueError("M must be a square power of 2")

        symbols = self._createConstellation(M)

        L = int(round(math.sqrt(M)))
        grayMappingIndexes = self._calculateGrayMappingIndexQAM(L)
        # noinspection PyUnresolvedReferences
        symbols = symbols[grayMappingIndexes]

        # Set the constellation
        self.setConstellation(symbols)

    @staticmethod
    def _createConstellation(M: int) -> np.ndarray:
        """
        Generates the Constellation for the (SQUARE) M-QAM modulation
        scheme.

        Parameters
        ----------
        M : int
            The modulation cardinality

        Returns
        -------
        symbols : np.ndarray
            The QAM constellation with the desired cardinality.
        """
        # Size of the square. The square root is exact
        symbols = np.empty(M, dtype=complex)
        L = int(round(math.sqrt(M)))
        for jj in range(0, L):
            for ii in range(0, L):
                symbol = complex(-(L - 1) + jj * 2, (L - 1) - ii * 2)
                symbols[ii * L + jj] = symbol

        average_energy = (M - 1) * 2.0 / 3.0
        # Normalize the constellation, so that the mean symbol energy is
        # equal to one.
        return symbols / math.sqrt(average_energy)

    @staticmethod
    def _calculateGrayMappingIndexQAM(L: int) -> np.ndarray:
        """
        Calculates the indexes that should be applied to the
        constellation created by _createConstellation in order to
        correspond to Gray mapping.

        Notice that the square M-QAM constellation is a matrix of dimension
        L x L, where L is the square root of M. Since the constellation was
        generated without taking into account the Gray mapping, then we
        need to reorder the generated constellation and this function
        calculates the indexes that can be applied to the original
        constellation in order to do exactly that.

        As an example, for the 16-QAM modulation the indexes can be
        organized (row order) in the matrix below

        ====  ======  ======  ======  ======
         /      00      01      11      10
        ====  ======  ======  ======  ======
         00    0000    0001    0011    0010
         01    0100    0101    0111    0110
         11    1100    1101    1111    1110
         10    1000    1001    1011    1010
        ====  ======  ======  ======  ======

        This is equivalent to concatenate a Gray mapping for the row with a
        Gray mapping for the column, and the corresponding indexes are
        [0, 1, 3, 2, 4, 5, 7, 6, 12, 13, 15, 14, 8, 9, 11, 10]

        Parameters
        ----------
        L : int
            Square root of the modulation cardinality (must be an integer).

        Returns
        -------
        indexes : np.ndarray
            indexes that should be applied to the constellation created by
            _createConstellation in order to correspond to Gray mapping

        """
        # Row vector with the column variation (second half of the index in
        # binary form)
        column = binary2gray(np.arange(0, L, dtype=int))

        # Column vector with the row variation
        #
        # Column vector with the row variation (first half of the index in
        # binary form)
        row = column.reshape(L, 1)
        columns = np.tile(column, (L, 1))
        rows = np.tile(row, (1, L))
        # Shift the first part by half the number of bits and sum with the
        # second part to form each element in the index matrix
        index_matrix = (rows << (level2bits(L**2) // 2)) + columns

        # Return the indexes as a vector (row order, which is the default
        # in numpy)
        return np.reshape(index_matrix, L**2)

    # noinspection PyPep8
    def _calcTheoreticalSingleCarrierErrorRate(
            self, SNR: NumberOrArray) -> NumberOrArray:
        """
        Calculates the theoretical (approximation) error rate of a single
        carrier in the QAM system (QAM has two carriers).

        Parameters
        ----------
        SNR : float | np.ndarray
            Signal-to-noise-value (in dB).

        Returns
        -------
        Psc : float | np.ndarray
            The theoretical single carrier error rate.

        Notes
        -----
        This method is used in the :meth:`calcTheoreticalSER`
        implementation.

        See also
        --------
        calcTheoreticalSER

        """
        snr = dB2Linear(SNR)
        # Probability of error of each carrier in a square QAM
        # $P_{sc} = 2\left(1 - \frac{1}{\sqrt M}\right)Q\left(\sqrt{\frac{3}{M-1}\frac{E_s}{N_0}}\right)$
        sqrtM = np.sqrt(self._M)
        Psc = (2. * (1. - (1. / sqrtM)) *
               qfunc(np.sqrt(snr * 3. / (self._M - 1.))))
        return Psc  # type: ignore

    def calcTheoreticalSER(self, SNR: NumberOrArray) -> NumberOrArray:
        """
        Calculates the theoretical (approximation) symbol error rate for
        the QAM scheme.

        Parameters
        ----------
        SNR : float | np.ndarray
            Signal-to-noise-value (in dB).

        Returns
        -------
        SER : float | np.ndarray
            The theoretical symbol error rate.
        """
        Psc = self._calcTheoreticalSingleCarrierErrorRate(SNR)
        # The SER is then given by
        # $ser = 1 - (1 - Psc)^2$
        ser = 1 - (1 - Psc)**2
        return ser

    def calcTheoreticalBER(self, SNR: NumberOrArray) -> NumberOrArray:
        """
        Calculates the theoretical (approximation) bit error rate for
        the QAM scheme.

        Parameters
        ----------
        SNR : float | np.ndarray
            Signal-to-noise-value (in dB).

        Returns
        -------
        BER : float | np.ndarray
            The theoretical bit error rate.
        """
        # For higher SNR values and gray mapping, each symbol error
        # corresponds to approximately a single bit error. The BER is then
        # given by the probability of error of a single carrier in the QAM
        # system divided by the number of bits transported in that carrier.
        k = level2bits(self._M)
        Psc = self._calcTheoreticalSingleCarrierErrorRate(SNR)
        ber = (2. * Psc) / k
        return ber


# xxxxx End of QAM Class xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
